/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */

 




/* GPL_HEADER */
package org.greenstone.gatherer.gui.metaaudit;
/**************************************************************************************
 * Title:        Gatherer
 * Description:  The Gatherer: a tool for gathering and enriching a digital collection.
 * Copyright:    Copyright (c) 2001
 * Company:      The University of Waikato
 * Written:        /05/02
 * Revised:      22/08/02 Revamped, Optimized and Commented.
 **************************************************************************************/

import java.util.ArrayList;
/** An Autofilter object stores the filters set on a single column, and provides a method for determining if a certain value passes the filter.
 * @author John Thompson
 * @version 2.3
 */
public class Autofilter {
    /** <i>true</i> if the filter should be applied, <i>false</i> to indicate the filter is turned off. */
    public boolean active;
    /** <i>true</i> if the matching for the first expression should be case sensitive, <i>false</i> otherwise. */
    public boolean casesense_one;
    /** <i>true</i> if the matching for the second expression should be case sensitive, <i>false</i> otherwise. */
    public boolean casesense_two;
    /** Used to determine the operation intended when applying two filters, and set using the values of the OPERATION_TYPE enumeration. */
    public boolean operation;
    /** Used to determine how the column this filter is applied to should be sorted, and set using the values of the SORT_TYPE enumeration. */
    public boolean sort;
    /** The method to be used for the first filter expression, set from the values of the METHOD_LIST enumeration. */
    public int method_one;
    /** The method to be used for the second filter expression, set from the values of the METHOD_LIST enumeration. */
    public int method_two;
    /** The value to be matched against for the first expression. */
    public String value_one;
    /** The value to be matched against for the second expression. */
    public String value_two;
    /** An element of the SORT_TYPE enumeration, indicates lowest to highest value column ordering. */
    public static final boolean ASCENDING = true;
    /** An element of the OPERATION_TYPE enumeration, indicates that both filter expressions must be met (conjunction). */
    public static final boolean AND = true;
    /** An element of the SORT_TYPE enumeration, indicates highest to lowest value column ordering. */
    public static final boolean DESCENDING = false;
    /** An element of the OPERATION_TYPE enumeration, indicates that either (or both) filter expressions must be met (disjunction). */
    public static final boolean OR = false;
    /** An enumeration of symbolic names of various matching methods. */
    public static final String METHOD_LIST[] = {"eqeq", "!eq", "<", "<eq", ">", ">eq", "^", "!^", "$", "!$", "?", "!?"};
    /** Default Constructor. */
    public Autofilter() {
	operation = OR;
	sort = ASCENDING;
    }
    /** Determine if this filter is currently active.
     * @return <i>true</i> if it is active, <i>false</i> otherwise.
     */
    public boolean active() {
	return active;
    }
    /** Determine if this list of values (for a certain cell) passes the filter.
     * @param values An <strong>ArrayList</strong> of values sourced from a single cell in the associated column.
     * @return <i>true</i> if the values match and should be displayed, <i>false</i> otherwise.
     */
    public boolean filter(ArrayList values) {
	boolean result = false;
	if(value_one != null) {
	    result = filter(values, method_one, value_one, casesense_one);
	    if(result) {
		if(operation == AND && value_two != null) {
		    result = filter(values, method_two, value_two, casesense_two);
		}
	    }
	    else if(operation == OR && value_two != null) {
		result = filter(values, method_two, value_two, casesense_two);
	    }
	}
	return result;
    }
    /** Set the current activity state of this filter.
     * @param active The new state of this filter, <i>true</i> to activate, <i>false</i> otherwise.
     */
    public void setActive(boolean active) {
	this.active = active;
    }
    /** Set one of the filter expressions using the given information.
     * @param number The number of the filter you wish to set as an <i>int</i>. Either 1 or 2.
     * @param method An <i>int</i> indicating the method to be used when matching.
     * @param value The <strong>String</strong> to be matched against.
     * @param casesense <i>true</i> if this expression should be case sensitive, <i>false</i> otherwise.
     */
    public void setFilter(int number, int method, String value, boolean casesense) {
	if(!casesense && value != null) {
	    value = value.toLowerCase();
	}
	if(number == 1) {
	    casesense_one = casesense;
	    method_one = method;
	    value_one = value;
	}
	else {
	    casesense_two = casesense;
	    method_two = method;
	    value_two = value;
	}
    }
    /** Set the operation to be used to join the two filters (if a second filter is set).
     * @param operation <i>true</i> for conjunct filters, <i>false</i> for disjunct.
     */ 
    public void setOperation(boolean operation) {
	this.operation = operation;
    }
    /** Set the sort order of this column.
     * @param sort <i>true</i> for ascending sort, <i>false</i> for descending.
     */
    public void setSort(boolean sort) {
	this.sort = sort;
    }
    /** Decide whether a row should be displayed or filtered. The result depends on the selector set with setFilterType.
     * @param values An <strong>ArrayList</strong> of values to be checked against the filter.
     * @param method The method of matching to be used, as an <i>int</i>.
     * @param target The <strong>String</Strong> to match against.
     * @param casesense <i>true</i> if the match is to be case sensitive, <i>false</i> otherwise.
     * @return <i>true</i> to display the row, <i>false</i> to hide it.
     */
    public boolean filter(ArrayList values, int method, String target, boolean casesense) {
	boolean result = false;
	// There are several special cases when the filter always returns turn, such as when the target is the wildcard character.
	if(target == null || target.length() == 0 || target.equals("*")) {
	    result = true;
	}
	else {
				// For each value in the list...
	    for(int i = 0; i < values.size(); i++) {
		boolean pass;
		String source;
		// Account for case sensitivity.
		if(casesense) {
		    source = values.get(i).toString();
		}
		else {
		    source = values.get(i).toString().toLowerCase();
		}
		// Perform the match, based on the selected method.
		switch(method) {
		case 1: // !EQ
		    pass = !(source.equals(target));
		    break;
		case 2: // <
		    pass = (source.compareTo(target) < 0);
		    break;
		case 3: // <eq
		    pass = (source.compareTo(target) <= 0);
		    break;
		case 4: // >
		    pass = (source.compareTo(target) > 0);
		    break;
		case 5: // >eq
		    pass = (source.compareTo(target) >= 0);
		    break;
		case 6: // ^
		    pass = source.startsWith(target);
		    break;
		case 7: // !^
		    pass = !(source.startsWith(target));
		    break;
		case 8: // $
		    pass = source.endsWith(target);
		    break;
		case 9: // !$
		    pass = !(source.endsWith(target));
		    break;
		case 10: // ?
		    pass = (source.indexOf(target) != -1);
		    break;
		case 11: // !?
		    pass = (source.indexOf(target) == -1);
		    break;
		default: // EQEQ
		    pass = source.equals(target);
		    break;
		}
		result = result || pass;
	    }
	}
	return result;		 
    }
    /** Produce a textual representation of this autofilter.
     * @return A <strong>String</strong> displaying details of this autofilter.
     */
    public String toString() {
	String result = "One: " + method_one + " - " + value_one + " - " + casesense_one;
	if(value_two != null) {
	    result = result + "\n" + "Operation: " + (operation?"AND":"OR") + "\n";
	    result = result + "Two: " + method_two + " - " + value_two + " - " + casesense_two;
	}
	return result;
    }
}








