/*
 *    GS2Retrieve.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.service;


// Greenstone classes
import org.greenstone.gdbm.*;
import org.greenstone.gsdl3.util.*;

// XML classes
import org.w3c.dom.Document;
import org.w3c.dom.Element; 
import org.w3c.dom.NodeList;

// General Java classes
import java.io.File;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Set;
import java.util.Iterator;

/** Implements the generic retrieval and classifier services for GS2 
 * collections.
 *
 * @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 * @author <a href="mailto:mdewsnip@cs.waikato.ac.nz">Michael Dewsnip</a>
 * @version $Revision: 4268 $
 */

public abstract class GS2Retrieve
    extends ServiceRack {

    // the services on offer
    // these strings must match what is found in the properties file
    protected static final String DOCUMENT_STRUCTURE_RETRIEVE_SERVICE = "DocumentStructureRetrieve";
    protected static final String DOCUMENT_METADATA_RETRIEVE_SERVICE = "DocumentMetadataRetrieve";
    protected static final String DOCUMENT_CONTENT_RETRIEVE_SERVICE = "DocumentContentRetrieve";


    // the browsing services -  now in here, these will only be advertised if classifiers have been specified in the config file
    private static final String CLASSIFIER_SERVICE = "ClassifierBrowse";
    private static final String CLASSIFIER_METADATA_SERVICE = "ClassifierBrowseMetadataRetrieve";
    
    protected static final String STRUCT_PARAM = "structure";
    protected static final String INFO_PARAM = "info";
    
    protected static final String STRUCT_ANCESTORS = "ancestors";
    protected static final String STRUCT_PARENT = "parent";
    protected static final String STRUCT_SIBS = "siblings";
    protected static final String STRUCT_CHILDREN = "children";
    protected static final String STRUCT_DESCENDS = "descendants";

    protected static final String INFO_NUM_SIBS = "numSiblings";
    protected static final String INFO_NUM_CHILDREN = "numChildren";
    protected static final String INFO_SIB_POS = "siblingPosition";

    protected static final int DOCUMENT=1;
    protected static final int CLASSIFIER=2;

    protected GDBMWrapper gdbm_src_ = null;
    
    protected Element config_info_ = null; // the xml from the config file

    /** constructor */
    protected GS2Retrieve()
    {
	gdbm_src_ = new GDBMWrapper();
    }


    /** configure this service */
    public boolean configure(Element info, Element extra_info)
    {
	System.out.println("Configuring GS2Retrieve...");
	config_info_ = info;

	// set up short_service_info_ - for now just has name and type
	Element dsr_service = doc_.createElement(GSXML.SERVICE_ELEM);
	dsr_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	dsr_service.setAttribute(GSXML.NAME_ATT, DOCUMENT_STRUCTURE_RETRIEVE_SERVICE);
	short_service_info_.appendChild(dsr_service);

	Element dmr_service = doc_.createElement(GSXML.SERVICE_ELEM);
	dmr_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	dmr_service.setAttribute(GSXML.NAME_ATT, DOCUMENT_METADATA_RETRIEVE_SERVICE);
	short_service_info_.appendChild(dmr_service);

	Element dcr_service = doc_.createElement(GSXML.SERVICE_ELEM);
	dcr_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
	dcr_service.setAttribute(GSXML.NAME_ATT, DOCUMENT_CONTENT_RETRIEVE_SERVICE);
	short_service_info_.appendChild(dcr_service);

	// set up service_info_map_ - for now, just has the same elements as above
	// should have full details about each service incl params lists etc.
	service_info_map_.put(DOCUMENT_STRUCTURE_RETRIEVE_SERVICE, dsr_service);
	service_info_map_.put(DOCUMENT_METADATA_RETRIEVE_SERVICE, dmr_service);
	service_info_map_.put(DOCUMENT_CONTENT_RETRIEVE_SERVICE, dcr_service);

	// Open GDBM database for querying
	String gdbm_db_file = GSFile.GDBMDatabaseFile(site_home_, cluster_name_);
	if (!gdbm_src_.openDatabase(gdbm_db_file, GDBMWrapper.READER)) {
	    System.err.println("Error: Could not open GDBM database!");
	    return false;
	}

        // now do the classifier browse service
 
        // check that there are classifiers specified
        Element class_list = (Element)GSXML.getChildByTagName(info, GSXML.CLASSIFIER_ELEM+GSXML.LIST_MODIFIER);
        if (class_list == null) {
            // no classifiers specified
            return true;
        }
	
	// get the display and format elements from the coll config file for
	// the classifiers
	extractExtraClassifierInfo(info, extra_info);
	config_info_ = info;
	
        // short_service_info_ - the browse one
        Element cb_service = doc_.createElement(GSXML.SERVICE_ELEM);
        cb_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_BROWSE);
        cb_service.setAttribute(GSXML.NAME_ATT, CLASSIFIER_SERVICE);
        short_service_info_.appendChild(cb_service);
	
	// metadata retrieval for the browsing  
        Element cbmr_service = doc_.createElement(GSXML.SERVICE_ELEM);
        cbmr_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_RETRIEVE);
        cbmr_service.setAttribute(GSXML.NAME_ATT, CLASSIFIER_METADATA_SERVICE);
        short_service_info_.appendChild(cbmr_service);

        // service_info_map_

	// the metadata one has the same info for now  - should this advertise what metadata is available??
	Element cbmr_service_info = (Element) cbmr_service.cloneNode(true);
	service_info_map_.put(CLASSIFIER_METADATA_SERVICE, cbmr_service_info);

	//the browse one
	Element cb_service_info = (Element)cb_service.cloneNode(true);
	Element cl_list = doc_.createElement(GSXML.CLASSIFIER_ELEM+GSXML.LIST_MODIFIER);
	cb_service_info.appendChild(cl_list);
	
	// the format info
	Element cb_format_info = doc_.createElement(GSXML.FORMAT_ELEM);
	boolean format_found = false;
	// add in to the description a simplified list of classifiers
        NodeList classifiers = class_list.getElementsByTagName(GSXML.CLASSIFIER_ELEM);
	for(int i=0; i<classifiers.getLength(); i++) {
	    Element cl = (Element)classifiers.item(i);
	    Element new_cl = (Element)doc_.importNode(cl, false); // just import this node, not the children
	    
	    cl_list.appendChild(new_cl);

	    // get the format info out, and put inside a classifier element
	    Element format_cl = (Element)new_cl.cloneNode(false);
	    Element format = (Element)GSXML.getChildByTagName(cl, GSXML.FORMAT_ELEM);
	    if (format != null) {
		
		//copy all the children
		NodeList elems = format.getChildNodes();
		for (int j=0; j<elems.getLength();j++) {
		    format_cl.appendChild(doc_.importNode(elems.item(j), true));
		}
		cb_format_info.appendChild(format_cl);
		format_found = true;
	    }
	    
	    
	}
	    
	
	service_info_map_.put(CLASSIFIER_SERVICE, cb_service_info);
	if (format_found) {
	    format_info_map_.put(CLASSIFIER_SERVICE, cb_format_info);
	}
	return true;
    }
    
    /** this looks for any classifier specific display or format info from extra_info and adds it in to the correct place in info */
    protected boolean extractExtraClassifierInfo(Element info, Element extra_info) {
	
	if (extra_info == null) {
	    return false;
	}
	
	Document owner = info.getOwnerDocument();
	// so far we have display and format elements that we need for classifiers
	NodeList classifiers = info.getElementsByTagName(GSXML.CLASSIFIER_ELEM);
	Element config_browse = (Element)GSXML.getChildByTagName(extra_info, GSXML.BROWSE_ELEM);

	for (int i=0; i<classifiers.getLength();i++) {
	    Element cl = (Element)classifiers.item(i);
	    String name = cl.getAttribute(GSXML.NAME_ATT);
	    Element node_extra = GSXML.getNamedElement(config_browse,
						       GSXML.CLASSIFIER_ELEM,
						       GSXML.NAME_ATT,
						       name);
	    if (node_extra == null) {
		System.err.println("GS2REtrieve: haven't found extra info for classifier named "+name);
		continue;
	    }
	    
	    // get the display elements if any - displayName
	    NodeList display_names = node_extra.getElementsByTagName(GSXML.DISPLAYNAME_ELEM);
	    if (display_names !=null) {
		Element display = owner.createElement(GSXML.DISPLAY_ELEM);
		for (int j=0; j<display_names.getLength(); j++) {
		    Element e = (Element)display_names.item(j);
		    
		    Element display_name = GSXML.createTextElement(owner, GSXML.DISPLAY_NAME_ELEM, GSXML.getNodeText(e));
		    display_name.setAttribute(GSXML.LANG_ATT, e.getAttribute(GSXML.LANG_ATT));
		    display.appendChild(display_name);
		}
		cl.appendChild(display);
	    }
	    
	    // get the format element if any 
	    Element format = (Element)GSXML.getChildByTagName(node_extra, GSXML.FORMAT_ELEM);
	    if (format==null) { // try a generic one that applies to all classifiers
		format = (Element)GSXML.getChildByTagName(extra_info,
							  GSXML.FORMAT_ELEM);
	    }
	    if (format!=null) { // append to index info
		cl.appendChild(owner.importNode(format, true));
	    }
	} // for each classifier
	return true;
    }
	    

    /** creates a display element containing all the text strings needed to display the service page, in the language specified 
     * the retrieval services dont get displayed to the users - they are only used internally by the actions. so this returns an empty display element
     * for those services. CLASSIFIER_BROWSE service returns some info */
    protected Element createServiceDisplay(String service, String lang)
    {
	Element display = doc_.createElement(GSXML.DISPLAY_ELEM);
	if (!service.equals(CLASSIFIER_SERVICE)) {
	    return display;
	}

	// CLASSIFIER_SERVICE
	display.appendChild(GSXML.createTextElement(doc_, GSXML.DISPLAY_NAME_ELEM,  getTextString(service+".name", lang)));
	//display.appendChild(GSXML.createTextElement(doc_,  GSXML.DISPLAY_SUBMIT_ELEM, getTextString(service+".submit", lang)));
	
	// need to add in the classifier name info
	// add <classifier name="CL1">text name</classifier> to the 
	// display node
	NodeList classifiers = config_info_.getElementsByTagName(GSXML.CLASSIFIER_ELEM);
	for (int i=0; i<classifiers.getLength(); i++) {
	    Element cl = (Element)classifiers.item(i);
	    Element disp = (Element)GSXML.getChildByTagName(cl, GSXML.DISPLAY_ELEM);
	    String text = null;
	    if (disp !=null) {
		text = GSXML.getDisplayText(disp, 
					    GSXML.DISPLAY_NAME_ELEM, 
					    lang, "en");
	    }
	    if (text == null || text.equals("")) {
		// no display element was specified, use the metadata name
		// for now this looks in the class properties file
		// this needs to use a general metadata thing instead
		text = getTextString(cl.getAttribute(GSXML.CLASSIFIER_CONTENT_ATT), lang);
	    }
	    
	    Element cl_elem = doc_.createElement(GSXML.CLASSIFIER_ELEM);
	    cl_elem.setAttribute(GSXML.NAME_ATT, cl.getAttribute(GSXML.NAME_ATT));
	    Element cl_name = GSXML.createTextElement(doc_, GSXML.DISPLAY_NAME_ELEM, text);
	    cl_elem.appendChild(cl_name);
		
	    display.appendChild(cl_elem);
	    
	}
	
	return display;
	
    }

    /** parent is true if this node is definitely the  parent of something, 
     * child is true is it definitely is a child of something - just for efficiency purposes */
    protected Element createDocNode(String doc_id, boolean parent, boolean child) {
	
	// create this here or pass it in?
	DBInfo db_info = gdbm_src_.getInfo(doc_id);
	Element node;
	if (isClassifier(doc_id)) {
	    node = doc_.createElement(GSXML.CLASS_NODE_ELEM);
	    String childtype = db_info.getInfo("childtype");
	    String orientation="";
	    if (childtype.equals("HList")) {
		orientation = "horizontal";
	    } else { // assume vertical
		orientation = "vertical";
	    }
	    node.setAttribute(GSXML.CLASS_NODE_ORIENTATION_ATT, orientation);
	} else {
	    node = doc_.createElement(GSXML.DOC_NODE_ELEM);
	    if (/*!child &&*/ OID.isTop(doc_id)) {
		node.setAttribute(GSXML.NODE_TYPE_ATT, GSXML.NODE_TYPE_ROOT);
	    } else if (!parent && isLeafNode(doc_id)) {
		node.setAttribute(GSXML.NODE_TYPE_ATT, GSXML.NODE_TYPE_LEAF);
	    } else {
		node.setAttribute(GSXML.NODE_TYPE_ATT, GSXML.NODE_TYPE_INTERIOR);
	    }
	}
	node.setAttribute(GSXML.NODE_ID_ATT, doc_id);
	return node;

    }
    /** Returns the parent of a specified documentID, or null if none exists */
    protected Element getParent(String doc_id)
    {
	String parent_id = OID.getParent(doc_id);
	if (parent_id.equals(doc_id))
	    return null;
	
	return createDocNode(parent_id, true, false);
    }


    /** adds all the children of doc_id the the doc element,
     * and if recursive=true, adds all their children as well*/
    protected void addDescendants(Element doc, String doc_id, 
				  boolean recursive)
    {
	DBInfo info = gdbm_src_.getInfo(doc_id);
	String contains = info.getInfo("contains");

	StringTokenizer st = new StringTokenizer(contains, ";");
	while (st.hasMoreTokens()) {
	    String child_id = st.nextToken().replaceAll("\"", doc_id);
	    Element child = createDocNode(child_id, false, true);
	    doc.appendChild(child);

	    // Apply recursively, if desired
	    if (recursive) {
		addDescendants(child, child_id, recursive);
	    }
	    
	}
    }

    /** adds all the siblings of current_id to the parent element. */
    protected Element addSiblings(Element parent, String parent_id, String current_id) {
	Element current_node = (Element)parent.getFirstChild();
	if (current_node ==null) {
	    // create a sensible error message
	    System.err.println("Error: there should be a first child.");
	    return null;
	}
	// remove the current child,- will add it in later in its correct place
	parent.removeChild(current_node);

	// add in all the siblings,
	addDescendants(parent, parent_id, false);

	// find the node that is now the current node
	// this assumes that the new node that was created is the same as 
	// the old one that was removed - we may want to replace the new one 
	// with the old one.
	Element new_current = GSXML.getNamedElement(parent, current_node.getNodeName(), GSXML.NODE_ID_ATT, current_id);
	return new_current;

    }
    /** Returns true if the OID specifies a leaf node, false otherwise
	Note: this makes a request to the GDBM database so it may not be
	      a particularly cheap operation */
    protected boolean isLeafNode(String oid)
    {
	DBInfo info = gdbm_src_.getInfo(oid);
	String children = info.getInfo("contains");
	return (children.equals(""));
    }

    // for now just use CL for classifiers - should have a type? in teh gdbm
    // database.
    protected boolean isClassifier(String oid) {
	if (oid.startsWith("CL")) {
	    return true;
	}
	return false;
    }

    protected Element processDocumentStructureRetrieve(Element request) {
	return genericStructureRetrieve(request, DOCUMENT);
    }

    protected Element processClassifierBrowse(Element request) {
	return genericStructureRetrieve(request, CLASSIFIER);
    }

    /** Retrieve the structure of a document */
    protected Element genericStructureRetrieve(Element request, int type)
    {
	// Create a new (empty) result message
	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);
	
	String node_name;
	String service_name;
	if (type==DOCUMENT) {
	    service_name = DOCUMENT_STRUCTURE_RETRIEVE_SERVICE;
	    node_name = GSXML.DOC_NODE_ELEM;
	} else {
	    service_name = CLASSIFIER_SERVICE;
	    node_name = GSXML.CLASS_NODE_ELEM;
	}
	
	result.setAttribute(GSXML.FROM_ATT, service_name);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_PROCESS);

	// Get the parameters of the request
	Element param_list = (Element) GSXML.getChildByTagName(request, GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	if (param_list == null) {
	    System.err.println("Error: DocumentStructureRetrieve request had no paramList.");
	    return result;  // Return the empty result
	}

	// the type of info required
	boolean want_structure = false;
	boolean want_info = false;

	Vector info_types=new Vector();
	// The document structure information desired
	boolean want_ancestors = false;
	boolean want_parent = false;
	boolean want_siblings = false;
	boolean want_children = false;
	boolean want_descendants = false;

	// Process the request parameters
	NodeList params = param_list.getElementsByTagName(GSXML.PARAM_ELEM);
	for (int i=0; i<params.getLength();i++) {
	    
	    Element param = (Element)params.item(i);
	    String p_name = param.getAttribute(GSXML.NAME_ATT);
	    String p_value = GSXML.getValue(param);
	    // Identify the structure information desired
	    if (p_name.equals(STRUCT_PARAM)) {
		want_structure = true;
		
		// This is NOT locale sensitive
		if (p_value.equals(STRUCT_ANCESTORS))
		    want_ancestors = true;
		else if (p_value.equals(STRUCT_PARENT))
		    want_parent = true;
		else if (p_value.equals(STRUCT_SIBS))
		    want_siblings = true;
		else if (p_value.equals(STRUCT_CHILDREN))
		    want_children = true;
		else if (p_value.equals(STRUCT_DESCENDS))
		    want_descendants = true;
		else
		    System.err.println("Warning: Unknown value \"" + p_value + "\".");
	    } else if (p_name.equals(INFO_PARAM)) {
		want_info = true;
		info_types.add(p_value);
	    }
	}

	// Make sure there is no repeated information
	if (want_ancestors)
	    want_parent = false;
	if (want_descendants)
	    want_children = false;
	
	
	Element query_doc_list = (Element) GSXML.getChildByTagName(request, node_name+GSXML.LIST_MODIFIER);
	if (query_doc_list == null) {
	    System.err.println("Error: DocumentStructureRetrieve request specified no doc nodes.\n");
	    return result;
	}
	
	Element doc_list = doc_.createElement(node_name+GSXML.LIST_MODIFIER);
	result.appendChild(doc_list);
	
	// Get the documents
	String[] doc_ids = GSXML.getAttributeValuesFromList(query_doc_list, 
							    GSXML.NODE_ID_ATT);
	for (int i = 0; i < doc_ids.length; i++) {
	    String doc_id = doc_ids[i];
	    
	    System.out.println("doc_id = "+doc_id);
	    if (OID.needsTranslating(doc_id)) {
		doc_id = gdbm_src_.translateOID(doc_id);
	    System.out.println("translated doc_id = "+doc_id);
	    }

	    // Add the document to the list
	    Element doc = doc_.createElement(node_name);
	    doc_list.appendChild(doc);
	    doc.setAttribute(GSXML.NODE_ID_ATT, doc_id);


	    if (want_info) {

		Element info_elem = doc_.createElement("nodeStructureInfo");
		doc.appendChild(info_elem);
		
		for (int j=0; j<info_types.size(); j++) {
		    String info_type = (String)info_types.get(j);
		    Element inf = getInfo(doc_id, info_type);
		    if (inf != null) {
			info_elem.appendChild(inf);
		    }
		}
	    }
	    if (want_structure) {
		// all structure info goes into a nodeStructure elem
		Element structure_elem = doc_.createElement(GSXML.NODE_STRUCTURE_ELEM);
		doc.appendChild(structure_elem);
		
		// Add the requested structure information
		Element current = createDocNode(doc_id, false, false);
		
		//Ancestors: continually add parent nodes until the root is reached
		Element top_node = current; // the top node so far
		if (want_ancestors) {
		    String current_id = doc_id;
		    while (true) {
			Element parent = getParent(current_id);
			if (parent == null)
			    break;
			
			parent.appendChild(top_node);
			current_id = parent.getAttribute(GSXML.NODE_ID_ATT);
			top_node = parent;
		    }
		}
		// Parent: get the parent of the selected node
		if (want_parent) {
		    Element parent = getParent(doc_id);
		    if (parent != null) {
			parent.appendChild(current);
			top_node = parent;
		    }
		}

		
		// now the top node is the root of the structure
		structure_elem.appendChild(top_node);
		
		//Siblings: get the other descendants of the selected node's parent
		if (want_siblings) {
		    Element parent = (Element)current.getParentNode(); // this may be the structure element if there has been no request for parents or ancestors
		    String parent_id = OID.getParent(doc_id);
		    
		    // add siblings, - returns a pointer to the new current node 
		    current = addSiblings(parent, parent_id, doc_id);
		}
		
	    // Children: get the descendants, but only one level deep
		if (want_children)
		    addDescendants(current, doc_id, false);
		// Descendants: recursively get every descendant of the selected node
		if (want_descendants)
		    addDescendants(current, doc_id, true);
	    } // if want structure
	} // for each doc
	return result;
    }


    protected Element processDocumentMetadataRetrieve(Element request) {
	return genericMetadataRetrieve(request, DOCUMENT);
    }

     protected Element processClassifierBrowseMetadataRetrieve(Element request) {
	return genericMetadataRetrieve(request, CLASSIFIER);
    }
   
    
    /** Retrieve metadata associated with a document or classifier node*/
    protected Element genericMetadataRetrieve(Element request, int type)
    {
	// Create a new (empty) result message
	Element result = doc_.createElement(GSXML.RESPONSE_ELEM);

	String node_name;
	
	String service_name;
	if (type==DOCUMENT) {
	    service_name = DOCUMENT_METADATA_RETRIEVE_SERVICE; 
	    node_name = GSXML.DOC_NODE_ELEM;
	} else {
	    service_name = CLASSIFIER_METADATA_SERVICE; 
	    node_name = GSXML.CLASS_NODE_ELEM;
	}
	result.setAttribute(GSXML.FROM_ATT, service_name);
	result.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_PROCESS);

	// Get the parameters of the request
	Element param_list = (Element) GSXML.getChildByTagName(request, GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	if (param_list == null) {
	    System.err.println("GS2Retrieve, DocumentMetadataRetrieve Error: missing paramList.\n");
	    return result;  // Return the empty result
	}

	// The metadata information required
	Vector metadata_list = new Vector();
	boolean all_metadata = false;
	// Process the request parameters
	Element param = (Element) param_list.getFirstChild();
	while (param != null) {
	    // Identify the metadata information desired
	    if (param.getAttribute(GSXML.NAME_ATT).equals("metadata")) {
		String metadata = GSXML.getValue(param);
		if (metadata.equals("all")) {
		    all_metadata = true;
		    break;
		}
		metadata_list.add(metadata);
	    }
	    param = (Element) param.getNextSibling();
	}
	
	Element node_list = doc_.createElement(node_name+GSXML.LIST_MODIFIER);
	result.appendChild(node_list);
	
	// Get the documents
	Element request_node_list = (Element) GSXML.getChildByTagName(request, node_name+GSXML.LIST_MODIFIER);
	if (request_node_list == null) {
	    System.err.println("Error: DocumentMetadataRetrieve request had no "+node_name+"List.\n");
 	    return result;
 	}
	
	NodeList request_nodes = request_node_list.getChildNodes();
	for (int i = 0; i < request_nodes.getLength(); i++) {
	    Element request_node = (Element) request_nodes.item(i);
	    String node_id = request_node.getAttribute(GSXML.NODE_ID_ATT);

	    if (OID.needsTranslating(node_id)) {
		node_id = gdbm_src_.translateOID(node_id);
	    }

	    // Add the document to the list
	    Element new_node = (Element)doc_.importNode(request_node, false);
	    node_list.appendChild(new_node);

	    // Add the requested metadata information
	    Element node_meta_list = doc_.createElement(GSXML.METADATA_ELEM+GSXML.LIST_MODIFIER);
	    new_node.appendChild(node_meta_list);
	    DBInfo info = gdbm_src_.getInfo(node_id);
	    if (info == null) {// I have had a case where it is null!
		continue;
	    }
	    System.out.println("found the info");
	    if (all_metadata) {
		System.out.println("trying to get all the metadata");
		// return everything out of the database
		Set keys = info.getKeys();
		Iterator it = keys.iterator();
		while(it.hasNext()) {
		    String key = (String)it.next();
		    System.out.println("getting metadata "+key);
		    String value = info.getInfo(key);
		    GSXML.addMetadata(doc_, node_meta_list, key, value);
		}
	    } else { // just get the selected ones
		
		for (int m = 0; m < metadata_list.size(); m++) {
		    String metadata = (String) metadata_list.get(m);
		    String value = info.getInfo(metadata);
		    GSXML.addMetadata(doc_, node_meta_list, metadata, value);
		}
	    }
	}
	
	return result;
    }


    /** Retrieve the content of a document - implemented by concrete subclasses */
    protected abstract Element processDocumentContentRetrieve(Element request);

    /** needs to get info from gdbm database - if the calling code gets it already it may pay to pass it in instead */
    protected String resolveImages(String doc_content, String doc_id)
    {
	String top_doc_id = OID.getTop(doc_id);
	DBInfo info = gdbm_src_.getInfo(top_doc_id);
	String archivedir = info.getInfo("archivedir");
	String image_dir  = site_http_address_ + "/collect/"+cluster_name_+"/index/assoc/"+archivedir;

	// Resolve all "_httpdocimg_"s
	doc_content = doc_content.replaceAll("_httpdocimg_", image_dir);
	return doc_content;
    }

    protected Element getInfo(String doc_id, String info_type) {

	String value="";
	if (info_type.equals(INFO_NUM_SIBS)) {
	    String parent_id = OID.getParent(doc_id);
	    if (parent_id.equals(doc_id)) {
		value="0";
	    } else {
		value = String.valueOf(getNumChildren(parent_id));
	    }
	} else if (info_type.equals(INFO_NUM_CHILDREN)) {
	    value = String.valueOf(getNumChildren(doc_id));
	} else if (info_type.equals(INFO_SIB_POS)) {
	    String parent_id = OID.getParent(doc_id);
	    if (parent_id.equals(doc_id)) {
		value="-1";
	    } else {
		DBInfo info = gdbm_src_.getInfo(parent_id);
		String contains = info.getInfo("contains");
		contains = contains.replaceAll("\"", parent_id);
		String [] children = contains.split(";");
		for (int i=0;i<children.length;i++) {
		    String child_id = children[i];
		    System.out.println("child="+child_id+" doc="+doc_id);
		    if (child_id.equals(doc_id)) {
			value = String.valueOf(i+1); // make it from 1 to length
			break;
		    }
		}
	    }
	} else {
	    return null;
	}
	Element info_elem = doc_.createElement("info");
	info_elem.setAttribute(GSXML.NAME_ATT, info_type);
	info_elem.setAttribute(GSXML.VALUE_ATT, value);
	return info_elem;
    }

    protected int getNumChildren(String doc_id) {
	DBInfo info = gdbm_src_.getInfo(doc_id);
	String contains = info.getInfo("contains");
	if (contains.equals("")) {
	    return 0;
	}
	String [] children = contains.split(";");
	return children.length;
    }
    
}   
